@@ -19,22 +19,23 @@
 
 """Provides the FocusTrackingPresenter for Orca."""
 
-__id__        = "$Id: focus_tracking_presenter.py,v 1.69 2006/06/16 14:05:56 wwalker Exp $"
-__version__   = "$Revision: 1.69 $"
-__date__      = "$Date: 2006/06/16 14:05:56 $"
+__id__        = "$Id: focus_tracking_presenter.py,v 1.84 2006/07/28 19:49:30 wwalker Exp $"
+__version__   = "$Revision: 1.84 $"
+__date__      = "$Date: 2006/07/28 19:49:30 $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
 import gobject
 import Queue
-#import threading
+import threading
 import time
 
 import atspi
+import braille
 import default
 import debug
 import input_event
-import orca
+import orca_state
 import presentation_manager
 import settings
 import speech
@@ -64,10 +65,16 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
         # is what delves them out, we keep at most one listener to avoid
         # receiving the same event twice in a row.
         #
-        self.registry      = atspi.Registry()
-        self._knownScripts = {}
-        self._eventQueue   = Queue.Queue(0)
-        self._gidle_id     = 0
+        self.registry        = atspi.Registry()
+        self._knownScripts   = {}
+        self._eventQueue     = Queue.Queue(0)
+        self._gidleId        = 0
+        self._gidleLock      = threading.Lock()
+        self.noFocusTimestamp = 0.0
+
+        if settings.debugEventQueue:
+            self._enqueueEventCount = 0
+            self._dequeueEventCount = 0
 
     ########################################################################
     #                                                                      #
@@ -174,7 +181,6 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                                       "...found %s.py" % name)
                         break
                     except ImportError:
-                        debug.printException(debug.LEVEL_ALL)
                         debug.println(debug.LEVEL_FINEST,
                                       "...could not find %s.py" % name)
                     except:
@@ -214,7 +220,6 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                 debug.println(debug.LEVEL_FINE,
                               "...found %s.py" % name)
             except ImportError:
-                debug.printException(debug.LEVEL_ALL)
                 debug.println(
                     debug.LEVEL_FINE,
                     "...could not find %s.py" % app.toolkitName)
@@ -346,7 +351,18 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
             return
 
         try:
+            # We don't want to touch a defunct object.  It's useless and it
+            # can cause hangs.
+            #
+            if event.source \
+               and event.source.state.count(atspi.Accessibility.STATE_DEFUNCT):
+                debug.println(debug.LEVEL_FINEST,
+                              "IGNORING DEFUNCT OBJECT")
+                atspi.Accessible.deleteAccessible(event.source)
+                return
+
             debug.printDetails(debug.LEVEL_FINEST, "    ", event.source)
+
         except CORBA.COMM_FAILURE:
             debug.printException(debug.LEVEL_WARNING)
             debug.println(debug.LEVEL_FINEST,
@@ -378,7 +394,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
         # as bugzilla bug 319777.]]]
         #
         retryCount = 0
-        oldLocusOfFocus = orca.locusOfFocus
+        oldLocusOfFocus = orca_state.locusOfFocus
         while retryCount <= settings.commFailureAttemptLimit:
             try:
                 if event.type == "window:activate":
@@ -406,7 +422,7 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
                     # of focus changed during our last attempt at
                     # handling this event.
                     #
-                    orca.locusOfFocus = oldLocusOfFocus
+                    orca_state.locusOfFocus = oldLocusOfFocus
                     debug.println(debug.LEVEL_WARNING,
                                   "  TRYING AGAIN (%d)" % retryCount)
                     time.sleep(settings.commFailureWaitTime)
@@ -437,6 +453,14 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
         #    self._processObjectEvent(atspi.Event(e))
         #return
 
+        if settings.debugEventQueue:
+            if self._enqueueEventCount:
+                debug.println(debug.LEVEL_ALL,
+                              "focus_tracking_presenter._enqueueEvent has " \
+                              "been entered before exiting (count = %d)" \
+                              % self._enqueueEventCount)
+            self._enqueueEventCount += 1
+
         event = None
         if isinstance(e, input_event.KeyboardEvent):
             if e.type == atspi.Accessibility.KEY_PRESSED_EVENT:
@@ -457,6 +481,8 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
             # care of them for us.
             #
             if (e.type == "object:state-changed:defunct"):
+                if settings.debugEventQueue:
+                    self._enqueueEventCount -= 1
                 return
 
             # We also generally do not like
@@ -466,6 +492,8 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
             # module take care of it for us.
             #
             if e.type == "object:property-change:accessible-parent":
+                if settings.debugEventQueue:
+                    self._enqueueEventCount -= 1
                 return
 
             # At this point in time, we only care when objects are
@@ -473,6 +501,8 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
             #
             if (e.type == "object:children-changed:remove") \
                 and (e.source != self.registry.desktop):
+                if settings.debugEventQueue:
+                    self._enqueueEventCount -= 1
                 return
 
             # We create the event here because it will ref everything
@@ -484,80 +514,152 @@ class FocusTrackingPresenter(presentation_manager.PresentationManager):
             event = atspi.Event(e)
 
         if event:
+            if settings.debugEventQueue:
+                debug.println(debug.LEVEL_ALL,
+                              "           acquiring lock...")
+            self._gidleLock.acquire()
+            if settings.debugEventQueue:
+                debug.println(debug.LEVEL_ALL,
+                              "           ...acquired")
+            if settings.debugEventQueue:
+                debug.println(debug.LEVEL_ALL,
+                              "           calling queue.put...")
+            if settings.debugEventQueue:
+                debug.println(debug.LEVEL_ALL,
+                              "           (full=%s)" % self._eventQueue.full())
             self._eventQueue.put(event)
-            if not self._gidle_id:
-                self._gidle_id = gobject.idle_add(self._dequeueEvent)
+            if settings.debugEventQueue:
+                debug.println(debug.LEVEL_ALL,
+                              "           ...put complete")
+            if not self._gidleId:
+                if settings.gilSleepTime:
+                    time.sleep(settings.gilSleepTime)
+                self._gidleId = gobject.idle_add(self._dequeueEvent)
 
-    def _timeout(self):
-        """Timer that will be called if we time out while trying to perform
-        an operation."""
-        debug.println(debug.LEVEL_SEVERE,
-                      "TIMEOUT: Looks like something has hung. Quitting Orca.")
-        orca.shutdown()
+            if settings.debugEventQueue:
+                debug.println(debug.LEVEL_ALL,
+                              "           releasing lock...")
+            self._gidleLock.release()
+            if settings.debugEventQueue:
+                debug.println(debug.LEVEL_ALL,
+                              "           ...released")
+        if settings.debugEventQueue:
+            self._enqueueEventCount -= 1
+
+        # [[[TODO: HACK - on some hangs, we see the event queue growing,
+        # but the thread to take the events off is hung.  We might be
+        # able to 'recover' by quitting when we see this happen.]]]
+        #
+        #if self._eventQueue.qsize() > 500:
+        #    print "Looks like something has hung.  Here's the threads:"
+        #    for someThread in threading.enumerate():
+        #        print someThread.getName(), someThread.isAlive()
+        #    print "Quitting Orca."
+        #    orca.shutdown()
 
     def _dequeueEvent(self):
         """Handles all events destined for scripts.  Called by the GTK
         idle thread.
         """
 
-        event = self._eventQueue.get()
+        rerun = True
 
-        if isinstance(event, input_event.KeyboardEvent):
-            if event.type == atspi.Accessibility.KEY_PRESSED_EVENT:
-                debug.println(debug.LEVEL_FINEST,
-                              "DEQUEUED KEYPRESS '%s' (%d) <----------" \
-                              % (event.event_string, event.hw_code))
-                pressRelease = "PRESS"
-            elif event.type == atspi.Accessibility.KEY_RELEASED_EVENT:
-                debug.println(debug.LEVEL_FINEST,
-                              "DEQUEUED KEYRELEASE '%s' (%d) <----------" \
-                              % (event.event_string, event.hw_code))
-                pressRelease = "RELEASE"
-            debug.println(debug.eventDebugLevel,
-                          "\nvvvvv PROCESS KEY %s EVENT %s vvvvv"\
-                          % (pressRelease, event.event_string))
-            self._processKeyboardEvent(event)
-            debug.println(debug.eventDebugLevel,
-                          "\n^^^^^ PROCESS KEY %s EVENT %s ^^^^^"\
-                          % (pressRelease, event.event_string))
-        elif isinstance(event, input_event.BrailleEvent):
-            debug.println(debug.LEVEL_FINEST,
-                          "DEQUEUED BRAILLE COMMAND %d <----------" \
-                          % event.event)
-            debug.println(debug.eventDebugLevel,
-                          "\nvvvvv PROCESS BRAILLE EVENT %d vvvvv"\
-                          % event.event)
-            self._processBrailleEvent(event)
-            debug.println(debug.eventDebugLevel,
-                          "\n^^^^^ PROCESS BRAILLE EVENT %d ^^^^^"\
-                          % event.event)
-        else:
-            debug.println(debug.LEVEL_FINEST, "DEQUEUED EVENT %s <----------" \
-                          % event.type)
+        if settings.debugEventQueue:
+            debug.println(debug.LEVEL_ALL,
+                          "Entering focus_tracking_presenter._dequeueEvent" \
+                          + " %d" % self._dequeueEventCount)
+            self._dequeueEventCount += 1
 
-            # [[[TODO: WDW - the timer stuff is an experiment to see if
-            # we can recover from hangs.  It's only experimental, so it's
-            # commented out for now.]]]
+        try:
+            event = self._eventQueue.get_nowait()
+
+            if isinstance(event, input_event.KeyboardEvent):
+                if event.type == atspi.Accessibility.KEY_PRESSED_EVENT:
+                    debug.println(debug.LEVEL_FINEST,
+                                  "DEQUEUED KEYPRESS '%s' (%d) <----------" \
+                                  % (event.event_string, event.hw_code))
+                    pressRelease = "PRESS"
+                elif event.type == atspi.Accessibility.KEY_RELEASED_EVENT:
+                    debug.println(debug.LEVEL_FINEST,
+                                  "DEQUEUED KEYRELEASE '%s' (%d) <----------" \
+                                  % (event.event_string, event.hw_code))
+                    pressRelease = "RELEASE"
+                debug.println(debug.eventDebugLevel,
+                              "\nvvvvv PROCESS KEY %s EVENT %s vvvvv"\
+                              % (pressRelease, event.event_string))
+                self._processKeyboardEvent(event)
+                debug.println(debug.eventDebugLevel,
+                              "\n^^^^^ PROCESS KEY %s EVENT %s ^^^^^"\
+                              % (pressRelease, event.event_string))
+            elif isinstance(event, input_event.BrailleEvent):
+                debug.println(debug.LEVEL_FINEST,
+                              "DEQUEUED BRAILLE COMMAND %d <----------" \
+                              % event.event)
+                debug.println(debug.eventDebugLevel,
+                              "\nvvvvv PROCESS BRAILLE EVENT %d vvvvv"\
+                              % event.event)
+                self._processBrailleEvent(event)
+                debug.println(debug.eventDebugLevel,
+                              "\n^^^^^ PROCESS BRAILLE EVENT %d ^^^^^"\
+                              % event.event)
+            else:
+                debug.println(debug.LEVEL_FINEST, "DEQUEUED EVENT %s <----------" \
+                              % event.type)
+
+                if (not debug.eventDebugFilter) \
+                    or (debug.eventDebugFilter \
+                        and debug.eventDebugFilter.match(event.type)):
+                    debug.println(debug.eventDebugLevel,
+                                  "\nvvvvv PROCESS OBJECT EVENT %s vvvvv" \
+                                  % event.type)
+                self._processObjectEvent(event)
+                if (not debug.eventDebugFilter) \
+                    or (debug.eventDebugFilter \
+                        and debug.eventDebugFilter.match(event.type)):
+                    debug.println(debug.eventDebugLevel,
+                                  "^^^^^ PROCESS OBJECT EVENT %s ^^^^^\n" \
+                                  % event.type)
+
+            # [[[TODO: HACK - it would seem logical to only do this if we
+            # discover the queue is empty, but this inroduces a hang for
+            # some reason if done inside an acquire/release block for a
+            # lock.  So...we do it here.]]]
             #
-            #timer = threading.Timer(5.0, self._timeout)
-            #timer.start()
+            noFocus = (not self._activeScript) \
+                      or ((not orca_state.locusOfFocus) \
+                          and (self.noFocusTimestamp \
+                               != orca_state.noFocusTimestamp))
 
-            debug.println(debug.eventDebugLevel,
-                          "\nvvvvv PROCESS OBJECT EVENT %s vvvvv" \
-                          % event.type)
-            self._processObjectEvent(event)
-            debug.println(debug.eventDebugLevel,
-                          "^^^^^ PROCESS OBJECT EVENT %s ^^^^^\n" \
-                          % event.type)
+            self._gidleLock.acquire()
+            if self._eventQueue.empty():
+                if noFocus:
+                    if settings.gilSleepTime:
+                        time.sleep(settings.gilSleepTime)
+                    message = _("No focus")
+                    if settings.brailleVerbosityLevel == \
+                        settings.VERBOSITY_LEVEL_VERBOSE:
+                        braille.displayMessage(message)
+                    if settings.speechVerbosityLevel == \
+                        settings.VERBOSITY_LEVEL_VERBOSE:
+                        speech.speak(message)
+                    self.noFocusTimestamp = orca_state.noFocusTimestamp
+                self._gidleId = 0
+                rerun = False # destroy and don't call again
+            self._gidleLock.release()
+        except Queue.Empty:
+            debug.println(debug.LEVEL_SEVERE,
+                          "focus_tracking_presenter:_dequeueEvent: " \
+                          + " the event queue is empty!")
+        except:
+            debug.printException(debug.LEVEL_SEVERE)
 
-            #timer.cancel()
-            #del timer
+        if settings.debugEventQueue:
+            self._dequeueEventCount -= 1
+            debug.println(debug.LEVEL_ALL,
+                          "Leaving focus_tracking_presenter._dequeueEvent" \
+                          + " %d" % self._dequeueEventCount)
 
-        if self._eventQueue.empty():
-            self._gidle_id = 0
-            return False # destroy and don't call again
-        else:
-            return True  # call again at next idle
+        return rerun
 
     def processKeyboardEvent(self, keyboardEvent):
         """Processes the given keyboard event based on the keybinding from the
